module m
  integer, parameter :: single = kind(1.)
  integer, parameter :: double = kind(1.d0)
  type matrix(K, N, M)
    integer, kind :: K; integer, len  :: N, M
    real(kind=K), dimension(N,M) :: a
  contains
    PROCEDURE, PRIVATE :: init_s
    PROCEDURE, PRIVATE :: init_d
    PROCEDURE, PRIVATE :: max_s => max__s
    PROCEDURE, PRIVATE :: max_d
    GENERIC :: max  => max_s,  max_d
    GENERIC :: init => init_s, init_d
  end type matrix
contains
  subroutine init_s( this )
    class(matrix(K=single, N=*, M=*)), intent(out) :: this
    call random_number( this%a )
  end subroutine init_s
  subroutine init_d( this )
    class(matrix(K=double, N=*, M=*)), intent(out) :: this
    call random_number( this%a )
  end subroutine init_d
  real(kind=single) function max__s( this )
    class(matrix(K=single, N=*, M=*)), intent(in) :: this
    max__s = MAXVAL( array=this%a )
  end function max__s
  real(kind=double) function max_d( this )
    class(matrix(K=double, N=*, M=*)), intent(in) :: this
    max_d = MAXVAL( array=this%a )
  end function max_d
end module m
program p
  use m
  type(matrix(K=single, N=10, M=20)) :: obj1
  type(matrix(K=double, N=20, M=50)) :: obj2
  real(kind=single) maxs; real(kind=double) maxd
  call obj1%init; call obj2%init
  maxs = obj1%max(); maxd = obj2%max()
  print *, "maxs = ", maxs 
  print *, "maxd = ", maxd 
end program p
