module fig
  use, intrinsic :: ISO_FORTRAN_ENV, only : OUTPUT_UNIT
  type, abstract :: figure_geo
    character(len=10) type
    contains
      procedure :: aire
      procedure :: volume
      procedure(aff), deferred :: affichage_constituants
  end type figure_geo

  type, extends(figure_geo) :: point
    private
    double precision x, y
    contains
      procedure :: affichage_constituants => affichage_constituants_point
  end type point
  type, extends(point) :: cercle
    private
    double precision rayon
    contains
      procedure :: aire                   => aire_cercle
      procedure :: affichage_constituants => affichage_constituants_cercle
  end type cercle

  type, extends(cercle) :: cylindre
    private
    double precision hauteur
    contains
      procedure :: aire                   => aire_cylindre
      procedure :: volume                 => volume_cylindre
      procedure :: affichage_constituants => affichage_constituants_cylindre
  end type cylindre

  type, extends(cercle) :: sphere
    contains
      procedure :: aire   => aire_sphere
      procedure :: volume => volume_sphere
  end type sphere

  interface point
    module procedure construct_point
  end interface point
  interface cercle
    module procedure construct_cercle
  end interface cercle
  interface cylindre
    module procedure construct_cylindre
  end interface cylindre
  interface sphere
    module procedure construct_sphere
  end interface sphere

  abstract interface
    subroutine aff( this )
      import figure_geo
      class(figure_geo), intent(in) :: this
    end subroutine aff
  end interface
contains

  ! Constructeurs
  function construct_point(x, y)
    double precision, intent(in) :: x, y
    type(point) :: construct_point

    construct_point%type = "POINT"
    construct_point%x = x; construct_point%y = y
  end function construct_point
  function construct_cercle(x, y, r)
    double precision, intent(in) :: x, y, r
    type(cercle) :: construct_cercle

    construct_cercle%point = point(x, y)
    construct_cercle%type = "CERCLE"
    construct_cercle%rayon = r
  end function construct_cercle
  function construct_cylindre(x, y, r, h)
    double precision, intent(in) :: x, y, r, h
    type(cylindre) :: construct_cylindre

    construct_cylindre%cercle = cercle(x, y, r)
    construct_cylindre%type = "CYLINDRE"
    construct_cylindre%hauteur = h
  end function construct_cylindre
  function construct_sphere(x, y, r)
    double precision, intent(in) :: x, y, r
    type(sphere) :: construct_sphere

    construct_sphere%cercle = cercle(x, y, r)
    construct_sphere%type = "SPHERE"
  end function construct_sphere
  function aire(this)
    class(figure_geo), intent(in) :: this
    double precision              :: aire

    aire = 0.d0
  end function aire
  function volume(this)
    class(figure_geo), intent(in) :: this
    double precision              :: volume

    volume = 0.d0
  end function volume
  function aire_cercle(this)
    class(cercle), intent(in) :: this
    double precision          :: aire_cercle

    aire_cercle = acos( -1.d0 )*this%rayon*this%rayon
  end function aire_cercle
  function aire_cylindre(this)
    class(cylindre), intent(in) :: this
    double precision            :: aire_cylindre

    aire_cylindre = 2.d0*this%cercle%aire() + &
                    2.d0*acos( -1.d0 )*this%rayon*this%hauteur
  end function aire_cylindre
  function aire_sphere(this)
    class(sphere), intent(in) :: this
    double precision          :: aire_sphere

    aire_sphere = 4.d0*this%cercle%aire()
  end function aire_sphere
  function volume_cylindre(this)
    class(cylindre), intent(in) :: this
    double precision            :: volume_cylindre

    volume_cylindre = this%cercle%aire()*this%hauteur
  end function volume_cylindre
  function volume_sphere(this)
    class(sphere), intent(in) :: this
    double precision          :: volume_sphere
    double precision          :: rayon

    rayon = this%cercle%rayon
    volume_sphere = 4.d0/3.d0*acos(-1.d0)*rayon*rayon*rayon
  end function volume_sphere
  subroutine affichage_constituants_point(this)
    class(point), intent(in) :: this

    write(OUTPUT_UNIT, '(2a,/,3(:,a,f0.2))')  &
                    "===> ", trim(this%type), &
                    "Coordonnees : [", this%x, ", ", this%y, "]"
  end subroutine affichage_constituants_point
  subroutine affichage_constituants_cercle(this)
    class(cercle), intent(in) :: this

    call this%point%affichage_constituants
    write(OUTPUT_UNIT, '(a, f10.2)') "Rayon       : ", this%rayon
  end subroutine affichage_constituants_cercle
  subroutine affichage_constituants_cylindre(this)
    class(cylindre), intent(in) :: this

    call this%cercle%affichage_constituants
    write(OUTPUT_UNIT, '(a, f10.2)') "Hauteur     : ", this%hauteur
  end subroutine affichage_constituants_cylindre
end module fig

program prog
  use fig

  type tab_ptr
    class(figure_geo), allocatable :: figure
  end type tab_ptr

  type(point)    p
  type(cercle)   c
  type(cylindre) cyl
  type(sphere)   sph

  p   = point   ( x=1.d0, y=5.d0 )
  c   = cercle  ( x=0.d0, y=5.d0, r=10.d0 )
  cyl = cylindre( x=0.d0, y=5.d0, r=10.d0, h=100.d0 )
  sph = sphere  ( x=0.d0, y=5.d0, r=10.d0 )

  call tableau
contains
  subroutine tableau
    type(tab_ptr), dimension(4) :: tab_figures
    integer n

    allocate( tab_figures(1)%figure, source=p )
    allocate( tab_figures(2)%figure, source=c )
    allocate( tab_figures(3)%figure, source=cyl )
    allocate( tab_figures(4)%figure, source=sph )
    do n=1,size(tab_figures,1)
      call sorties_figure(tab_figures(n)%figure)
      deallocate(tab_figures(n)%figure)
    end do
  end subroutine tableau
  !
  subroutine sorties_figure(f)
    use, intrinsic :: ISO_FORTRAN_ENV, only : OUTPUT_UNIT
    CLASS(figure_geo), intent(in) :: f

    call f%affichage_constituants
    write(OUTPUT_UNIT, '(a, f10.2)') "Aire        : ", f%aire()
    write(OUTPUT_UNIT, '(a, f10.2)') "Volume      : ", f%volume()
  end subroutine sorties_figure
end program prog
